import time
import sys
import os
import runpy
import processing
import threading

class HighlighterMonitor(processing.Process):

    def __init__(self, main_pipe):

        #Call the parent constructor
        processing.Process.__init__(self)
        
        #Store the main pipe which is used to communicate with the main program
        self.main_pipe = main_pipe

        #Initialise other variables
        self.is_paused = False
        self.slowing_enabled = False
        self.slowing_duration = 0
        self.do_step = False
        self.page_pipes = {}
        self.watched_files = []
        self.commit_suicide = False
        self.stdout_pipe = None
        self.stdin_pipe = None
        self.filename = __file__.replace(".pyc", ".py")
        
    def do_thread_stuff(self):
        while self.isAlive():
            self.process_messages()

    def process_messages(self):
        if self.main_pipe.poll(0.1):
            msg = self.main_pipe.recv()
            if msg == "D": #Die command
                self.commit_suicide = True
            elif msg == "N": #New pae
                new_file = self.main_pipe.recv()
                new_pipe = self.main_pipe.recv()
                self.add_page_pipe(new_pipe, new_file)
            elif msg == "P?": #Paused query
                self.main_pipe.send(self.is_paused)
            elif msg == "R": #Resume command
                self.resume()
            elif msg == "P": #Pause command
                self.pause()
            elif msg == "SSD": #Set Slowing Duration
                num = self.main_pipe.recv()
                self.set_slowing_duration(num)
            elif msg == "ST": #STep
                self.step()
            elif msg == "S?":
                self.main_pipe.send(self.slowing_enabled)
            elif msg == "SE": #Slowing Enable
                self.enable_slowing()
            elif msg == "SD": #Slowing Disable
                self.disable_slowing()

    def add_page_pipe(self, pipe, file_name):
        if file_name in self.page_pipes.keys():
            self.page_pipes[file_name].append(pipe)
        else:
            self.page_pipes.update({file_name:[pipe]})
            self.watched_files.append(file_name)

    def set_module_to_run(self,module_path):
        last_slash = module_path.rindex(os.sep)
        if ".py" in module_path:
            dotpy = module_path.rindex(".py")
        else:
            dotpy = None
        module_dir = module_path[:last_slash]
        module_name = module_path[last_slash+1:dotpy]

        sys.path.insert(0,module_dir)
    
        self.module_to_run = module_name

    def pause(self):
        """Pauses the execution of this thread by enabling a sleep loop in the tracer function."""
        if not self.is_paused:
            self.is_paused = True

    def resume(self):
        """Resumes from pause."""
        if self.is_paused:
            self.is_paused = False

    def enable_slowing(self):
        self.slowing_enabled = True

    def disable_slowing(self):
        self.slowing_enabled = False

    def set_slowing_duration(self, slow_time):
        self.slowing_duration = slow_time

    def step(self):
        if self.is_paused:
            self.do_step = True

    def set_stdout_pipe(self,pipe):
        self.stdout_pipe = pipe
    
    def set_stdin_pipe(self,pipe):
        self.stdin_pipe = pipe

    def run(self):
        #console hack: there are definitely cases where this will cause a program to behave differently until this is implemented better
        if self.stdout_pipe != None:
            class new_stdout():
                def __init__(myself):
                    try:
                        myself.closed = self.stdout_pipe.closed
                        myself.encoding = sys.stdout.encoding
                        myself.mode = sys.stdout.mode
                        myself.name = sys.stdout.name
                        myself.newlines = sys.stdout.newlines
                        myself.softspace = sys.stdout.softspace
                    except:
                        pass
                def close(*args):
                    self.closed = True
                    return self.stdout_pipe.close()
                def fileno(*args):
                    return sys.stdout.fileno()
                def flush(*args):
                    return
                def isatty(*args):
                    return sys.stdout.isatty()
                def next(*args):
                    return "Not supported by PyLighter."
                def read(*args):
                    return "Not supported by PyLighter."
                def readinto(*args):
                    return
                def readline(*args):
                    return "Not supported by PyLighter.\n"
                def readlines(*args):
                    return ["Not supported by PyLighter.\n"]
                def seek(*args):
                    return
                def tell(*args):
                    return 0
                def truncate(*args):
                    return
                def write(*args):
                    if self.stdout_pipe.closed:
                        return
                    self.stdout_pipe.send(str(args[1]))
                def writelines(*args):
                    if self.stdout_pipe.closed:
                        return
                    for line in args[1]:
                        self.stdout_pipe.send(str(line))
                def xreadlines(*args):
                    return
            processing.sys.stdout = processing.sys.stderr = new_stdout()
            
        if self.stdin_pipe != None:
            class new_stdin():
                def __init__(myself):
                    try:
                        myself.closed = self.stdin_pipe.closed
                        myself.encoding = sys.stdin.encoding
                        myself.mode = sys.stdin.mode
                        myself.name = sys.stdin.name
                        myself.newlines = sys.stdin.newlines
                        myself.softspace = sys.stdin.softspace
                    except:
                        pass
                def close(*args):
                    self.closed = True
                    return self.stdout_pipe.close()
                def fileno(*args):
                    return sys.stdout.fileno()
                def flush(*args):
                    return
                def isatty(*args):
                    return sys.stdout.isatty()
                def next(*args):
                    return "Not supported by PyLighter."
                def read(*args):
                    if self.stdin_pipe.closed:
                        return ""
                    return self.stdin_pipe.recv()
                def readinto(*args):
                    return
                def readline(*args):
                    if self.stdin_pipe.closed:
                        return ""
                    return self.stdin_pipe.recv()
                def readlines(*args):
                    result = []
                    if self.stdin_pipe.closed:
                        return []
                    while self.stdin_pipe.poll(0.001):
                        result.append(str(self.stdin_pipe.recv))
                    return result
                def seek(*args):
                    return
                def tell(*args):
                    return 0
                def truncate(*args):
                    return
                def write(*args):
                    return
                def writelines(*args):
                    return
                def xreadlines(*args):
                    return
            processing.sys.stdin = new_stdin()
        
        #start the thread to handle messages
        tee = threading.Thread(target=self.do_thread_stuff, name="PyLighterMonitorMessageThread")
        tee.start()
        
        #Define the tracer function to register for this thread
        def trace_function(frame, event, arg):
            if self.commit_suicide:
                exit(0)
        
            if event == "line":
                line_num = frame.f_lineno 
                try:
                    filename = frame.f_globals["__file__"]
                    if filename.endswith(".pyc"):
                        filename = filename.replace(".pyc",".py")
                except:
                    filename = "<Unknown file>"
                    
                while self.is_paused:
                    if filename == self.filename:
                        break
                    if self.do_step:
                        break
                    self.process_messages()
                    time.sleep(0.1)
                    
                try:
                    for pipe in self.page_pipes[filename]:
                        if pipe.poll(0):
                            msg = pipe.recv()
                            if msg == "rm":
                                self.page_pipes[filename].remove(pipe)
                        else:
                            pipe.send(line_num-1) #f_lineno gives back 1-based line numbers, so we need to adjust
                except KeyError:
                    pass
                else:
                    if self.slowing_enabled:
                        time.sleep(self.slowing_duration)
                    if self.do_step:
                        self.do_step = False
            return trace_function #A trace function is supposed to return itself
        
        sys.settrace(trace_function)
        threading.settrace(trace_function)

        time.sleep(0.1) #Wait just a moment for everything underneath to finish before starting the target
        
        try:
            runpy.run_module(self.module_to_run,run_name="__main__",alter_sys=True)
        except Exception, detail:
            import wx
            a = wx.App()
            wx.MessageDialog(None, caption="An error has occurred!", message="Whoops! There was an error in the running program!\n\nHere's Python's error message:\n\n%s" % detail, style=wx.ICON_ERROR|wx.OK).ShowModal()
            exit(1)
        
        while not self.commit_suicide:
            time.sleep(0.1)
